unit xbitmap_unit;

{
     -------------------------------------
   | DavData Software                     |
   | http://home.hccnet.nl/david.dirkse/  |
   | david.math@hccnet.nl                 |
   ---------------------------------------

 version 1.0
 27-01-2009

 additions to the bitmap class
 features:
           - fast direct acces of image memory
           - clipping rectangle
           - multi pixel wide dash-dot lines
           - 4 drawing levels [0..3] for pen and brush
           - improved stretchdraw method
}

interface

uses windows,graphics,dialogs,sysutils;//dialogs,sysutils;

type  PDW = ^DWORD;
      TXFillMapSize = (fms8,fms24);
      TXstretchmode = (smFullSource,smFullDest);
      TXBitmap = class(TBitmap)
  private
   FXpenWidth   : byte;
   FXLineStyle  : byte;//dash-dot pattern
   FXarrowCode  : byte;
   FXpencolor   : DWORD;//32 bit format --RGB
   FXpenlevel   : byte;
   FXbrushColor : DWORD;//32 bit format
   FXbrushlevel : byte;
   FXcliprect   : Trect;
   FXfillsize   : TXFillMapSize;//100bpi ..300bpi
   FXfillstyle  : byte;
   FXUseBrush   : boolean;
   FXStretchMode : TXstretchMode;
   FXModRect    : TRect;
//
   FXfillmap8    : array[0..7] of byte;
   FXfillMap24   : array[0..23] of LongInt;
   FXpen         : array[0..31] of DWORD; //pen image, "1" = paint
   FXlinePattern : Word;
   FXpenBias     : byte;//position of pen dot center
   FXPbase       : DWORD;//pointer to bitmap [0,0]
   FXlineStep    : integer;//nr of bytes between rows
//
   procedure fastVline(x1,y1,y2,color : integer);
   procedure fastHline(x1,y1,x2,color : integer);
   procedure line(x1,y1,x2,y2 : integer); //no arrows, penstyle
   procedure pijldimensions(var lengte, offset: single);
   procedure arc1(x1,y1,x2,y2,x3,y3,x4,y4 : integer; acode : byte);
   procedure makeModRect(l,t,r,b : integer;expans : byte);
   procedure polygon1(var pa : array of Tpoint);
   procedure Dot(x,y : integer);
  protected
   procedure setWidth(w : integer);override;
   procedure setHeight(h : integer);override;
   procedure setXClipRect(r : Trect);
   procedure setPenwidth(w : byte);
   procedure setXLinestyle(ps : byte);
   procedure setXPencolor(color : DWORD);
   function getXpencolor : DWORD;
   procedure setXpenlevel(lev : byte);
   procedure setXBrushColor(color : DWORD);
   function getXbrushColor : DWORD;
   procedure setXbrushlevel(lev : byte);
   procedure setXFillStyle(fst : byte);
   function getXpixel(x,y : integer):DWORD;
   procedure setXpixel(x,y : integer; dw : DWORD);
   function getPixPos(x,y : integer):DWORD;
  public
   constructor create;override;
   property  Xpixel[x:integer; y:integer]:DWORD read getXpixel write setXpixel;
   property  XPixPos[x:integer;y:integer]:DWORD read getPixPos;
   procedure Xdot(x,y : integer);
   procedure Xline(x1,y1,x2,y2 : integer);
   procedure Xellipse(x1,y1,x2,y2 : integer);
   procedure Xarc(x1,y1,x2,y2,x3,y3,x4,y4 : integer);
   procedure Xpolygon(var pa : array of Tpoint);
   procedure XfillRect(x1,y1,x2,y2 : integer; r : byte);
   procedure Xfloodfill(x1,y1 : integer);
   procedure XstretchRect(drt : Trect; sbm : TXBitmap;srt : Trect);
   procedure XAdjust;
   property  XpenWidth : byte read FXpenwidth write setpenWidth;
   property  Xpencolor : DWORD read getXpencolor write setXpencolor;
   property  XLineStyle : byte read FXLineStyle write setXlineStyle;
   property  Xbrushcolor : DWORD read getXbrushColor write setXBrushColor;
   property  Xpenlevel : byte read FXpenlevel write setXpenlevel;
   property  Xbrushlevel : byte read FXbrushlevel write setXbrushlevel;
   property  XFillMapSize : TXFillMapsize read FXfillSize write FXfillSize;
   property  XfillStyle : byte read FXFillStyle write setXFillStyle;
   property  XarrowCode : byte read FXarrowCode write FXarrowCode;
   property  XClipRect : Trect read FXclipRect write setXclipRect;
   property  XUseBrush : boolean read FXUseBrush write FXUseBrush;
   property  Xstretchmode : TXstretchmode read FXstretchmode write FXstretchmode;
   property  XModRect : Trect read FXModRect;
  end;

function UniRect(var r1, r2 : TRect) : Trect;
function InterRect(var r1,r2 : Trect) : Trect;
function PackRect(x1,y1,x2,y2 : integer) : Trect;
function swapcolor(c : DWORD) : DWORD;

implementation
                                                     //preset
const linepattern : array[0..19] of word =           //dash - dot patterns
                    ($ffff,$aaaa,$cccc,$eeee,$f0f0,
                     $f8f8,$fcfc,$fefe,$8888,$c0c0,
                     $e0e0,$ff00,$8080,$bebe,$9c9c,
                     $dede,$ff18,$ff88,$ffe4,$fffa);

//------- create -----------

constructor TXBitmap.create;
begin
 inherited create;
 width := 10;
 height := 10;
 FXpenwidth := 0;
 XpenWidth := 1;    //--<>-- width creates new image
 FXLineStyle := 0;
 FXlinePattern := linePattern[0];
 FXpenlevel := 0;
 FXbrushlevel := 3;
 FXpencolor := $000000;
 FXBrushColor := $ffffff;
 FXUseBrush := true;
 FXstretchMode := smFullSource;
 XFillMapSize := fms8;
 FXFillStyle := 100;
 XFillStyle := 0;
 FXArrowCode := 0;
 with FXModRect do
  begin
   left := -1; right := -1; top := -1; bottom := -1;
  end;
end;

procedure TXBitmap.XAdjust;
//call after change in bm size (create, load image)
begin
 pixelformat := pf32bit;
 if (width > 0) then
  begin
   if (height > 0) then
    begin
     FXPbase := DWORD(scanline[0]);
     with FXclipRect do
      begin
       left := 0; right := width;
       top := 0; bottom := height;
      end;
    end;
   if (height > 1) then FXlinestep := FXPBase - DWORD(scanline[1])
    else FXlinestep := 0;
  end;
end;

procedure TXBitmap.setWidth(w : integer);
begin
 inherited setwidth(w);
 Xadjust;
end;

procedure TXBitmap.setheight(h : integer);
begin
 inherited setHeight(h);
 XAdjust;
end;

procedure TXBitmap.pijldimensions(var lengte,offset: single);
//set length , offset of arrow at end of line or arc
const r : array[0..3] of single = (2.5 , 1 , 0.5 , 0.25);
begin
 offset := ((FXarrowCode and $3)+1)*((FXpenwidth shr 1)+1)+2;
 lengte := (offset+1)/ r[(FXArrowCode and $c) shr 2];
end;

//------------ private procedures -------------

procedure TXBitmap.fastVline(x1,y1,y2,color : integer);
//draw horizontal line, for filling shapes,rectangles
//only solid lines, penwidth = 1
//y1 > y2
//color is 32 bit format
var i : integer;
    pRow : DWORD;
    p : PDW;
    level : byte;
begin
 if (x1 < FXcliprect.left) or (x1 >= FXcliprect.right) then exit;
//--
 level := color and $3;
 if y2 < y1 then
  begin
   i := y1; y1 := y2; y2 := i;
  end; 
 pRow := FXpbase + (x1 shl 2) - y1*FXlineStep;
 for i := y1 to y2 do
  begin
   if (i >= FXcliprect.top) and (i < FXcliprect.bottom) then
    begin
     p := PDW(pRow);
     if level <= (p^ and $3) then p^ := color;
    end;
   dec(pRow,Fxlinestep);
  end;
end;

procedure TXBitmap.fastHline(x1,y1,x2,color : integer);
//draw horizontal line, for filling shapes,rectangles
//only solid lines penwidth = 1
//x1 < x2 , bit 0,1 of color is the level
//color is 32 bit format
var i : integer;
    pColumn : DWORD;
    p : PDW;
    level : byte;
begin
 if (y1 < FXcliprect.top) or (y1 >= FXcliprect.bottom) then exit;
//--
 level := color and $3;
 if x2 < x1 then
  begin
   i := x1; x1 := x2; x2 := i;
  end;
 pColumn := FXpbase-FXlineStep*(y1)+ (x1 shl 2);
 for i := x1 to x2 do
  begin
   if (i >= FXcliprect.left) and (i < FXcliprect.right) then
    begin
     p := PDW(pColumn);
     if level <= (p^ and $3) then p^ := color;
    end;
   inc(pColumn,4);
  end;
end;

//------------ Xdrawing -----------------

procedure TXBitmap.XDot(x,y : integer);
begin
 Dot(x,y);
 makeModRect(x,y,x,y,(FXPenwidth shr 1)+2);
end;

procedure TXBitmap.Dot(x,y : integer);
//write a single dot,
//use FXpenwidth,Xcliprect,Xlevel,FXpencolor
var i,j : byte;
    bitmask,pline : DWORD;
    p : PDW;
    xi,yj : integer;
begin
 x := x - FXpenBias;
 y := y - FXpenBias;
 for j := 0 to FXpenwidth-1 do
  begin
   bitmask := 1;
   yj := y + j;
   if (yj < FXcliprect.top) or (yj >= FXcliprect.bottom) then continue;
   pline := FXpbase-FXlineStep*(yj);
   for i := 0 to FXpenWidth-1 do
    begin
     xi := x + i;
     if (xi >= FXcliprect.left) and (xi < FXcliprect.right) then
      begin
       p := PDW(pline + (xi) shl 2);
       if (FXpen[j] and bitmask) <> 0 then
        if FXpenlevel <= (p^ and $3) then p^ := FXpencolor;
      end;
     bitmask := bitmask shl 1;
    end;//for i
  end;//for j
end;

procedure TXBitmap.line(x1,y1,x2,y2 : integer);
//paint line from (x1,y1) to (x2,y2);
//use xpenwidth,xlevel,cliprect  (no penstyle)
var dx,dy,x,y : single;
    steps,i : word;
begin
 dx := x2-x1; dy := y2 - y1;
 if (dx = 0) and (dy = 0) then
  begin
   Dot(x1,y1);
   exit;
  end;

 if abs(dy) <= abs(dx) then
  begin                                   //hor. orientation
   steps := abs(trunc(dx)); dy := dy/steps;
   if dx > 0 then dx := 1 else dx := -1;
  end
 else begin                               //vert orientation
       steps := abs(trunc(dy)); dx := dx/steps;
       if dy > 0 then dy := 1 else dy := -1;
      end;
 x := x1; y := y1;
 for i := 0 to steps do
  begin
   Dot(trunc(x+0.5),trunc(y+0.5));
   x := x + dx;
   y := y + dy;
  end;
end;

procedure TXBitmap.Xline(x1,y1,x2,y2 : integer);
//paint line from (x1,y1) to (x2,y2);
//use xpenwidth,xpenstyle,xlevel,cliprect
var dx,dy,ipen,x,y,w,PL,PD,PBX,PBY : single;
    steps,i : word;
    beginpijl,eindpijl : boolean;
    PA,PB : array[0..2] of TPoint;
    PLX,PLY : smallInt;

begin
 dx := x2-x1; dy := y2 - y1;
 if (dx = 0) and (dy = 0) then
  begin
   XDot(x1,y1);     //make modrect
   exit;
  end;

 beginpijl := (FXArrowCode and $40) <> 0;  //begin-arrow
 eindpijl := (FXArrowCode and $80) <> 0;   //end-arrow
 if beginpijl or eindpijl then
  begin
   w := sqrt(dx*dx+dy*dy);
   pijldimensions(PL,PD);                //get arrow dimensions PL, PD
   eindpijl := eindpijl and (1.2*PL < w);
   beginpijl := beginpijl and (2.2*PL < w);
   w := 1/w;
   PLX := trunc(PL*dx*w+0.5);
   PLY := trunc(PL*dy*w+0.5);            //arrow length
   PBX := PD*dy*w;
   PBY := PD*dx*w;                       //arrow width
  end;

 i := (FXPenwidth shr 1)+ 2;
 if (beginPijl or EindPijl) then inc(i,trunc(PD+1));
 makeModRect(x1,y1,x2,y2,i);

 if beginpijl then
  begin
   PA[0].x := x1; x1 := x1 + PLX;
   pa[1].x := trunc(x1+PBX+0.5);
   pa[2].x := trunc(x1-PBX+0.5);
   pa[0].y := y1; y1 := y1 + PLY;
   pa[1].y := trunc(y1-PBY+0.5);
   pa[2].y := trunc(y1+PBY+0.5);
  end;
 if eindpijl then
  begin
   pb[0].x := x2; x2 := x2-PLX;
   pb[1].x := trunc(x2 + PBX+0.5);
   pb[2].x := trunc(x2 - PBX+0.5);
   pb[0].y := y2; y2 := y2 - PLY;
   pb[1].y := trunc(y2 - PBY+0.5);
   pb[2].y := trunc(y2 + PBY+0.5);
  end;

 dx := x2-x1; dy := y2 - y1;              //must recalculate dx,dy
 if abs(dy) <= abs(dx) then
  begin                                   //hor. orientation
   steps := abs(trunc(dx)); dy := dy/steps;
   if dx > 0 then dx := 1 else dx := -1;
  end
 else begin                               //vert orientation
       steps := abs(trunc(dy)); dx := dx/steps;
       if dy > 0 then dy := 1 else dy := -1;
      end;
 ipen := 1/fxpenwidth;
 x := x1; y := y1;
 for i := 0 to steps do
  begin
   if ((($8000 shr (trunc(i*ipen) and $f)) and FXlinePattern) <> 0) and
      ((($8000 shr (trunc(i*ipen + 0.9) and $f)) and FXlinePattern) <> 0) then
      Dot(x1,y1);
   x := x + dx; x1 := trunc(x + 0.5);
   y := y + dy; y1 := trunc(y + 0.5);
  end;

 if beginpijl then Polygon1(pa);
 if eindpijl then Polygon1(pb);
end;

procedure TXBitmap.Xellipse(x1,y1,x2,y2 : integer);
//use penwidth,pencolor,penlevel,linestyle to draw ellipse
var mx,my,x,y,a,b : integer;
    b2,dab : single;
begin
 if FXuseBrush then
  begin
   mx := (x1+x2) div 2;
   my := (y1+y2) div 2;
   a := abs(x2-x1) shr 1;
   b := abs(y2-y1) shr 1;
   if (a = 0) or (b = 0) or (a > 1000) or (b > 1000) then exit;
//
   b2 := b*b;
   dab := a/b;
   for y := -b to b do
    begin                                  //background color
     x := trunc(dab*sqrt(b2 - y*y));
     fastHline(mx-x,my+y,mx+x,FXbrushColor);
    end;
  end;//if brush
 arc1(x1,y1,x2,y2,x1,y1,x1,y1,0);         //ellipse, no arrows
end;

procedure TXBitmap.XArc(x1,y1,x2,y2,x3,y3,x4,y4 : integer);
begin
 arc1(x1,y1,x2,y2,x3,y3,x4,y4,FXArrowcode);
end;

procedure TXBitmap.Arc1(x1,y1,x2,y2,x3,y3,x4,y4 : integer; acode : byte);
//common code for Xellipse,Xarc
//acode = 0 for  ellipse, = FXArrowcode for Arcs
//draw arc inside rect(x1,y1)..(x2,y2)
//from intersection with line M..(x3,y3)
//to intersection with line M..(x4,y4)  ...counterclockwise
//use penwidth, pencolor, penlevel
var a,b,a2,b2,mx,my,x,y,pa : integer;
    beginside, endside : byte;
    ab2,s,ipen,v1,v2,dab,dba : single;
    sp,ep : integer;                   //start-,endpoint
    beginpijl,eindpijl : boolean;      //begin arrow, endarrow (arcs)
    PL,PD : single;                    //arrow length, arrow width/2
    p1,p2 : array[0..2] of TPoint;     //arrow polygons
    pt : TPoint;
    sine,cosine,rc : single;           //arrow position calculation

//-- arrow support

    function wdist(p : integer; side : byte) : word;
    //calculate distance from righttop (x2,y2) to p on side
    begin
     case side of
      0 : result := x2 - p;
      1 : result := 2*x2 + p - y1;
      2 : result := 2*x2 + 2*y2 + p - x1;
      3 : result := 4*x2 + 2*y2 + y2 - p;
      else result := 0;
     end;//case
    end;

    procedure nextside(var p : integer; var side : byte);//p : point, s : side
    //use points x1,x2,y1,y2 as reference ,
    //walk (x2,y1)..(x1,y1)..(x1,y2)..(x2,y2)..(x2,y1) side 0..3
    begin
     case side of
      0 : if p = x1 then begin
                          p := y1; side := 1; nextside(p,side);
                         end;
      1 : if p = y2 then begin
                          p := x1; side := 2; nextside(p,side);
                         end;
      2 : if p = x2 then begin
                          p := y2; side := 3; nextside(p,side);
                         end;
      3 : if p = y1 then begin
                          p := x2; side := 0; nextside(p,side);
                         end;
     end;//case
    end;

    procedure nextpoint(var p : integer; var side : byte);
    //calculate next point and side
    begin
     if (side = 0) or (side = 3) then dec(p) else inc(p);
     nextside(p,side);
    end;

    procedure previousside(var p : integer; var side : byte);//p : point, s : side
    //use points x1,x2,y1,y2 as reference ,
    begin
     case side of
      0 : if p > x2 then begin
                          p := y1+1; side := 3; previousside(p,side);
                         end;
      1 : if p < y1 then begin
                          p := x1+1; side := 0; previousside(p,side);
                         end;
      2 : if p < x1 then begin
                          p := y2-1; side := 1; previousside(p,side);
                         end;
      3 : if p > y2 then begin
                          p := x2-1; side := 2; previousside(p,side);
                         end;
     end;//case
    end;

    procedure previouspoint(var p : integer; var side : byte);
    //calculate previous point and side
    begin
     if (side = 0) or (side = 3) then inc(p) else dec(p);
     previousside(p,side);
    end;

// -------------------------

    function WtoP(p : integer; side : byte) : TPoint;
    //point p on side s to point of ellipse
    //use a2,b2,dba,dab
    begin
     case side of
      0 : begin
           result.x := p;
           result.y := -trunc(dba*sqrt(a2-p*p)+0.25);
          end;
      1 : begin
           result.x := -trunc(dab*sqrt(b2-p*p)+0.25);
           result.y := p;
          end;
      2 : begin
           result.x := p;
           result.y := trunc(dba*sqrt(a2-p*p)+0.25);
          end;
      3 : begin
           result.x := trunc(dab*sqrt(b2-p*p)+0.25);
           result.y := p;
          end;
     end;//case
    end;

// ------------------------- main ------

begin
 a := (FXPenwidth shr 1)+ 2;
 if (acode and $c0) <> 0 then
  begin
   pijldimensions(PL,PD);
   inc(a,trunc(PD+1));
  end;
 makeModRect(x1,y1,x2,y2,a);

 mx := (x1+x2) div 2;       //sort x,y
 my := (y1+y2) div 2;
 a := abs(x2-x1) shr 1;
 b := abs(y2-y1) shr 1;
 if (a = 0) or (b = 0) or (a > 1000) or (b > 1000) then exit;
//--
 ipen := 1/fxpenwidth;   //penwidth divider

//--- calculate (x1,y1)....(x2,y2) of inner rectangle

 a2 := a*a;
 b2 := b*b;
 ab2 := sqrt(a2 + b2);
 x2 := trunc(a2/ab2 + 0.5);
 x1 := -x2;
 y2 := trunc(b2/ab2 + 0.5);
 y1 := -y2;
//
 if (x1 = 0) and (y1 = 0) then exit;
//
 x3 := x3 - mx; x4 := x4 - mx;
 y3 := y3 - my; y4 := y4 - my;
 if ((x3=0) and (y3=0)) then x3 := x2;
 if ((x4=0) and (y4=0)) then x4 := x2;
//
 v1 := b*x3;
 v2 := a*y3;
 s := a*b/sqrt(v1*v1 + v2*v2);
 x := trunc(x3*s + 0.5);
 y := trunc(y3*s + 0.5);
//
 dba := b/a;       //for WtoP procedure
 dab := a/b;
//
 p1[0].x := x + mx; p1[0].y := y + my;
 if (abs(x) <= x2) then                  //startpoint
  begin
   if (y3<=0) then beginside := 0 else beginside := 2;
   sp := x;
  end
 else begin
       if (x3<0) then beginside := 1 else beginside := 3;
       sp := trunc(y3*s+0.5);
      end;
 v1 := b*x4;                             //endpoint
 v2 := a*y4;
 s := a*b/sqrt(v1*v1 + v2*v2);
 x := trunc(x4*s + 0.5);
 y := trunc(y4*s + 0.5);
 p2[0].x := x + mx; p2[0].y := y + my;
 if abs(x) <= x2 then
  begin
   if (y4<=0) then endside := 0 else endside := 2;
   ep := x;
  end
 else  begin
        if (x4<0) then endside := 1 else endside := 3;
        ep := trunc(y4*s+0.5);
       end;
 nextside(sp,beginside);
 nextside(ep,endside);

//--  arrows -----------

 beginpijl := (ACode and $40) <> 0;  //begin-arrow
 eindpijl := (ACode and $80) <> 0;   //end-arrow
 if beginpijl or eindpijl then
  begin
   pa := wdist(ep,endside) - wdist(sp,beginside);
   if pa <= 0 then pa := 4*(x2 + y2) - pa;
   eindpijl := eindpijl and (1.2*PL < pa);
   beginpijl := beginpijl and (2.2*PL < pa);
   if beginpijl then
    begin
     for pa := 0 to trunc(PL + 0.5) do nextpoint(sp,beginside);
     pt := WtoP(sp,beginside);
     if pt.x = 0 then begin
                       sine := PD; cosine := 0;
                      end
     else begin
           rc := (pt.y * a2)/(pt.x * b2);
           sine := 1/sqrt(1 + rc*rc);
           cosine := sine * PD;
           sine := sine * rc * PD;
          end;
     P1[1].x := trunc(mx + pt.x + cosine + 0.5);
     P1[1].y := trunc(my + pt.y + sine + 0.5);
     P1[2].x := trunc(mx + pt.x - cosine + 0.5);
     P1[2].y := trunc(my + pt.y - sine + 0.5);
    end;
   if eindpijl then
    begin
     for pa := 0 to trunc(PL + 0.5) do previouspoint(ep,endside);
     pt := WtoP(ep,endside);
     if pt.x = 0 then begin                   
                       sine := PD; cosine := 0;
                      end
     else begin
           rc := (pt.y * a2)/(pt.x * b2);
           sine := 1/sqrt(1 + rc*rc);
           cosine := sine * PD;
           sine := sine * rc * PD;
          end;
     P2[1].x := trunc(mx + pt.x + cosine + 0.5);
     P2[1].y := trunc(my + pt.y + sine + 0.5);
     P2[2].x := trunc(mx + pt.x - cosine + 0.5);
     P2[2].y := trunc(my + pt.y - sine + 0.5);
    end;
  end;//if beginpijl or endpijl

// ------- draw --------------

 pa := 0;          //path length for dash-dot
 repeat
  pt := WtoP(sp,beginside);
  if ((($8000 shr (trunc(pa*ipen) and $f)) and FXlinePattern) <> 0) and
     ((($8000 shr (trunc(pa*ipen + 0.9) and $f)) and FXlinePattern) <> 0) then
        Dot(mx+pt.x,my+pt.y);
  nextPoint(sp,beginside);
  inc(pa);//total path
 until (beginside = endside) and (sp = ep);

 if beginpijl then polygon1(p1);
 if eindpijl then polygon1(p2);
end;

procedure TXBitmap.Xpolygon(var pa : array of Tpoint);
var i : byte;
    x1,y1,x2,y2 : integer;
begin
 polygon1(pa);                  //handle the polygon
 x1 := pa[0].x; x2 := pa[0].x;  //make modrect
 y1 := pa[0].y; y2 := pa[0].y;
 for i := 1 to high(pa) do
  begin
   if pa[i].x < x1 then x1 := pa[i].x;
   if pa[i].x > x2 then x2 := pa[i].x;
   if pa[i].y < y1 then y1 := pa[i].y;
   if pa[i].y > y2 then y2 := pa[i].y;
  end;
 i:= (XPenWidth shr 1) + 2;
 makeModRect(x1,y1,x2,y2,i);
end;

procedure TXBitmap.polygon1(var pa : array of Tpoint);
//paint polygon, connect first- with last point of pa
//if brush then color inside with brushcolor
type Tvector = record
                x,y,dx,dy : integer;
               end;
var lastP : shortInt;
    i,j,nexti : byte;
    n : integer;
    vlist : array[1..40] of Tvector;
    f1,f2 : single;
    valid,b1 : boolean;

label loop1,loop2,end1;

//---------support

  procedure vrsect(v1,v2 : TVector);
  //calculate intersection of vectors v1,v2
  //line1 = (v1.x1,v1.y1) +f1*(v1.dx,v1.dy)
  //line2 = (v2.x1,v2.y1) +f2*(v2.dx,v2.dy)
  //return f1,f2,valid
  var d,vx,vy : integer;
  begin
   d := v1.dx*v2.dy - v1.dy*v2.dx;//discriminant
   if d = 0 then begin
                  valid := false; exit;
                 end;
   valid := true;
   vx := v2.x - v1.x;
   vy := v2.y - v1.y;
   f1 := (vx*v2.dy - vy*v2.dx)/d;
   f2 := (vx*v1.dy - vy*v1.dx)/d;
  end;

  function Adjacent(a,b : byte): boolean;
  //return true if vectors Vlist[a],Vlist[b] share start- endpoint
  begin
   result := ((a = 1) and (b = n)) or
             ((b = 1) and (a = n)) or
             (abs(a-b) < 2);
  end;

  function outward(iv : byte) : boolean;
  //return true if  vlist[iv] points outward
  var ii : byte;
  begin
   result := false;
   for ii := 1 to n do
    begin
     if (ii <> iv) then
      begin
       vrsect(vlist[iv],vlist[ii]);
       if valid and (f1 > 1) and (f2 >= 0) and (f2 <= 1) then exit;
      end;
    end;//for
   result := true;
  end;

  function Innerpoint3(i1,i2 : byte) : boolean;
  //check for point inside triangle i1,i2 : false
  //no point : true
  var i3,k  : byte;
      vv,vp : TVector;
  begin
   i3 := i2 + 1;
   if i3 > n then i3 := 1;
   with vv do
    begin
     x := vlist[i1].x; y := vlist[i1].y;
     dx := vlist[i1].dx + vlist[i2].dx;
     dy := vlist[i1].dy + vlist[i2].dy;
    end;
   result := false;
   for k := 1 to n do
    begin
     if (k <> i1) and (k <> i2) and (k <> i3) then
      begin
       with vp do
        begin
         x := vlist[i2].x; y := vlist[i2].y;
         dx := vlist[k].x - x; dy := vlist[k].y - y;
        end;
       vrsect(vp,vv);
       if valid and (f1 >= 1) and (f2 >= 0) and (f2 <=1) then exit;
      end;
    end;
   result := true;
  end;

  procedure triColor(v1,v2 : Tvector);
  //color triangle with FXBrushColor
  //v1 must point to start of v2
  var h : integer;
      a : array[1..3] of TPoint;
      x1,y1,x2,y2,x3,y3 : integer;
      dx1,dx2,dx3,x,xx : single;
      ii,jj : byte;
  begin
   a[1].x := v1.x; a[1].y := v1.y;               //load array
   a[2].x := v2.x; a[2].y := v2.y;
   a[3].x := v2.x+v2.dx; a[3].y := v2.y+v2.dy;
   for ii := 1 to 2 do                           //sort array y - incr.
    for jj := ii+1 to 3 do
     if a[ii].y > a[jj].y then
      begin
       h := a[ii].x; a[ii].x := a[jj].x; a[jj].x := h;
       h := a[ii].y; a[ii].y := a[jj].y; a[jj].y := h;
      end;
   x1 := a[1].x; y1 := a[1].y;                  //load sorted points
   x2 := a[2].x; y2 := a[2].y;
   x3 := a[3].x; y3 := a[3].y;

   if y1 <> y3 then dx3 := (x3-x1)/(y3-y1);
   if y1 <> y2 then
    begin
     dx1 := (x2-x1)/(y2-y1);
     xx := x1;
    end
   else xx := x2;
   if y2 <> y3 then dx2 := (x3-x2)/(y3-y2);
   x := x1;
   for h := y1 to y3 do
    begin
     fastHline(trunc(x+0.5),h,trunc(xx+0.5),FXbrushColor);
     x := x + dx3;
     if h >= y2 then xx := xx + dx2 else xx := xx + dx1;
    end;
  end;

  procedure drawEdges;
  //draw line around shape
  var ii : byte;
  begin
   for ii := 0 to lastP-1 do line(pa[ii].x,pa[ii].y,pa[ii+1].x,pa[ii+1].y);
   line(pa[lastP].x,pa[lastP].y,pa[0].x,pa[0].y);
  end;

//----------- main Xpolygon1 ------------

begin
 lastP := high(pa);
 n := lastP+1;
 if (n < 3) or (n > 40) then exit;
 if XuseBrush = false then goto end1;

 i := 1;
 for j := 0 to lastP-1 do         //copy points to vectorlist
  with vlist[i] do
   begin
    x := pa[j].x; y := pa[j].y;
    dx := pa[j+1].x - x; dy := pa[j+1].y - y;
    inc(i);
   end;
  with vlist[i] do
   begin
    x := pa[lastP].x; y := pa[lastP].y;
    dx := pa[0].x - x; dy := pa[0].y - y;
   end;

 for i := 1 to n do                    //check duplicate points
  for j := i+1 to n do
   if (vlist[i].x = vlist[j].x) and (vlist[i].y = vlist[j].y) then
    begin
     drawEdges;
     exit;
    end;

 for i := 1 to n do        //check for no intersection
  for j := 1 to n do
   if adjacent(i,j) = false then
    begin
     vrsect(vlist[i],vlist[j]);
     if valid and ((f1 >= 0) and (f1 <= 1)) and ((f2 >= 0) and (f2 <= 1))then
      begin
       drawEdges;
       exit;
      end;
    end;
    
// ------  shape OK, no duplicate points ---- make triangles ------

loop1:

 i := 1;

loop2:

 nexti := i + 1;
 if nexti > n then nexti := 1; //correct overflow
 b1 := outward(i);
 if b1 then b1 := innerpoint3(i,nexti);
 if b1 then
  begin
   tricolor(vlist[i],vlist[nexti]);
   with vlist[i] do              //replace vlist[i] by vlist[i]+vlist[nexti]
    begin
     dx := dx + vlist[nexti].dx;
     dy := dy + vlist[nexti].dy;
    end;
   for j := nexti to n-1 do vlist[j] := vlist[j+1];//eliminate vlist[nexti]
   dec(n);
   if n >= 3 then goto loop1 else goto end1;
  end
 else
  begin
   inc(i);
   if (i > n) then goto end1;
   goto loop2;
  end;

end1:

 drawedges;

end;

procedure TXBitmap.XfillRect(x1,y1,x2,y2 : integer; r : byte);
//fill rct with selected brush color
// do not replace pixels with higher level then brushlevel
//r = radius for round angles. Set r = 0 for sharp rectangle angles
var i,i2,r2,x,y,delta : integer;
begin
 begin                        //sort x,y params
  if x1 > x2 then begin
                    r2 := x1; x1 := x2; x2 := r2;
                  end;
  if y1 > y2 then begin
                    r2 := y1; y1 := y2; y2 := r2;
                  end;

   i := (FXPenwidth shr 1) + 2;//modrect
   makeModRect(x1,y1,x2,y2,i);

   if 2*r >= x2 - x1 then r := (x2 - x1) shr 1;
   if 2*r >= y2 - y1 then r := (y2 - y1) shr 1;
   r2 := sqr(r);

   if fxUseBrush then              //fill
    begin
     for i := y1 to y2 do fastHline(x1+r,i,x2-r,fxbrushcolor);
     for i := 0 to r - 1 do
      begin
       i2 := sqr(r-i);
       delta := r - trunc(sqrt(r2 - i2)+0.5);
       fastVline(x1+i,y1+delta,y2-delta,fxbrushcolor);
       fastVline(x2-i,y1+delta,y2-delta,fxbrushcolor);
      end;
    end;//if
                                        //straight edges
   for i := 0 to Xpenwidth-1 do
    begin
     fastHline(x1+r,y1+i-FXpenbias,x2-r,FXpenColor);
     fastHline(x1+r,y2+i-FXpenbias,x2-r,FXpencolor);
     fastVline(x1-FXpenBias+i,y1+r,y2-r,FXpencolor);
     fastVline(x2+i-FXpenbias,y1+r,y2-r,FXpenColor);
    end;//for i

   for i := 0 to trunc(0.707*r + 0.5) do  //round edges
    begin
     i2 := sqr(i);
     delta := trunc(sqrt(r2 - i2)+0.5);
     x := x1 + r;
     y := y1 + r;
     dot(x-i,y-delta);                 //left top
     dot(x-delta,y-i);
     y := y2 - r;
     dot(x-i,y+delta);                 //left bottom
     dot(x-delta,y+i);
     x := x2 - r;
     dot(x+i,y+delta);                 //right bottom
     dot(x+delta,y+i);
     y := y1 + r;
     dot(x+i,y-delta);               //right top
     dot(x+delta,y-i);
    end;
  end;//with rct
end;

procedure TXBitmap.Xfloodfill(x1,y1 : integer);
//floodfill Xcliprect area starting at (x1,y1)
//boundaries are pixels with level 0
//leftmost byte of pixel is control
//------xx : direction 0..3
//---x---- : scanned   $10
//--x----- : fill      $20
//----x--- : start scan $8
//colortype is bits 0,1 of color, --00 = boundary
//brushcolor is used for painting
var p     : PDW;
    p1,cc : DWORD;
    x,y,dy : integer;
    b,dir,modx,mody : byte;
label nextpixel,previouspixel,fill;
begin
 with XClipRect do
  begin

//-----------check valid startpoint -------------

   if (x1 < left) or (x1 >= right) or (y1 < top) or (y1 >= bottom) then exit;

//------- mark boundaries -------------

   p1 := Fxpbase-Fxlinestep*top;
   for x := left to right-1 do
    begin
     p := PDW(p1 + (x shl 2));
     p^ := p^ or $10000000;//hor top line, mark bounderies
    end;
   p1 := Fxpbase-Fxlinestep*(bottom-1);
   for x := left to right-1 do
    begin
     p := PDW(p1 + (x shl 2));
     p^ := p^ or $10000000;//hor bottom line
    end;
   p1 := Fxpbase+(left shl 2);
   for y := top+1 to bottom-2 do
    begin
     p := PDW(p1 - y * FXlinestep);
     p^ := p^ or $10000000;//vert left line
    end;
   p1 := Fxpbase+((right-1) shl 2);
   for y := top+1 to bottom-2 do
    begin
     p := PDW(p1 - y * Fxlinestep);
     p^ := p^ or $10000000;//vert right line
    end;

//--------- mark colorlevel 0 --------

 for y := top+1 to bottom-2 do
  begin
   p1 := Fxpbase - y*FXlinestep;
   for x:= left+1 to right-2 do
    begin
     p := PDW(p1 + (x shl 2));
     cc := p^ and $ffffff;
     if (cc and $3) = 0 then cc := cc or $10000000;
     p^ := cc;
    end;
  end;
 end;//cliprect

//------------ scan area ------------

  p := PDW(Fxpbase - y1*FXlinestep + (x1 shl 2));
  cc := p^;//
  if (cc and $10000000) <> 0 then goto fill;//exit, if started on border
//
  dy := width;          //pointer decrement of line
  p^ := cc or $38000000;//start,fill,scanned bit
  dir := 0;

nextpixel:

    case dir of
     0 : inc(p);   //inc x
     1 : dec(p);   //dec x
     2 : inc(p,dy);//dec y
     3 : dec(p,dy);//inc y
    end;//case

    cc := p^;
    if (cc and $10000000) <> 0 then goto previouspixel//if scanned
     else
      begin
       cc := cc or (($30 or dir) shl 24); //mark scanned,fill
       p^ := cc;                          //store entry direction
      end;
    if dir <> 1 then dir := 0;  //make exit direction
    goto nextpixel;

previouspixel:

   case dir of
    0 : dec(p);   //dec x
    1 : inc(p);   //inc x
    2 : dec(p,dy);//inc y
    3 : inc(p,dy);//dec y
   end;
   inc(dir);
   b := (p^ shr 24) and $0f;
   if b = (dir xor 1) then inc(dir);//skip entry-direction
   if dir > 3 then
    begin
     dir := b;
     if dir >= 8 then goto fill else goto previouspixel;
    end
   else goto nextpixel;

//---------- fill --------------

fill:

 if xfillmapsize = fms24 then            //24 bit fill pattern
 with XCliprect do                       //for printer
  begin
   mody := top mod 24;
   for y := top to bottom-1 do
    begin
     p1 := FXpBase - y*FXlinestep;
     modx := left mod 24;
     for x := left to right-1 do
      begin
       p := PDW(p1 + (x shl 2));
       cc := p^;
       if ((cc and $20000000) <> 0) and
          ((Fxfillmap24[mody] and (1 shl modx)) <> 0) then cc := FXbrushcolor
        else cc := cc and $ffffff;
       p^:= cc;
       inc(modx);
       if modx = 24 then modx := 0;
      end;//for x
     inc(mody);
     if mody = 24 then mody := 0;
    end;//for y
  end//with xcliprect
 else                                   //8 bit fill pattern
   with XCliprect do                    //for screen
  begin
   for y := top to bottom-1 do
    begin
     p1 := FXpBase - y*FXlinestep;
     for x := left to right-1 do
      begin
       p := PDW(p1 + (x shl 2));
       cc := p^;
       if ((cc and $20000000) <> 0) and
           (Fxfillmap8[y and $7] and (1 shl (x and $7)) <> 0) then
             cc := FXbrushcolor
        else cc := cc and $ffffff;
       p^:= cc;
      end;//for x
    end;//for y
  end//with xcliprect
end;

//----------- stretchrect ----------------------

function floor(a : single) : word;
//return a rounded down to integer
begin
 result := trunc(a);
end;

function ceiling(a : single) : word;
//return a rounded up to integer
begin
 result := trunc(a);
 if frac(a) > 0.001 then inc(result);
end;

procedure TXBitmap.XstretchRect(drt:Trect; sbm:TXBitmap;srt:Trect);
//stretch sbm rectangle srt to fit drt of XBitmap
var sx1,sy1,sx2,sy2 : single;    //source field positions
    x,y,i,j : word;              //source,dest field pixels
    destR,destG,destB : single;  //destination colors
    sR,sG,sB : byte;             //source colors
    f,fi2 : single;              //factors
    dx,dy,AP : single;           //distance, area percentage
    color : DWord;//longInt;
    drwidth,drheight : word;
begin
 f := abs(drt.right-drt.left)/abs(srt.Right-srt.left);
 fi2 := abs(drt.bottom-drt.top)/abs(srt.bottom-srt.top);
 if FXstretchMode = smFullSource then
  begin
   if f > fi2 then f := fi2;                 //select smallest factor
   drwidth := trunc(abs(srt.right-srt.left)*f);
   drheight := trunc(abs(srt.bottom-srt.top)*f);
  end
  else
   begin
    if f < fi2 then f := fi2;             //select largest factor
    drwidth := abs(drt.right-drt.left);
    drheight := abs(drt.bottom-drt.top);
   end;
 fi2 := f * f;
 f := 1/f;                           //f = source / dest dimensions
//---
 for y := 0 to drheight-1 do         //vertical destination pixels
  begin
   sy1 := f * y;
   sy2 := sy1 + f;
   for x := 0 to drwidth-1 do        //horizontal destination pixels
    begin
     sx1 := f * x;
     sx2 := sx1 + f;
     destR := 0; destG := 0; destB := 0;       //clear colors
     for j := floor(sy1) to ceiling(sy2)-1 do  //vertical source pixels
      begin
       dy := 1;
       if sy1 > j then begin
                        dy := dy-(sy1-j);
                       end;
       if sy2 < j+1 then begin
                          dy := dy-(j+1-sy2);
                         end;
       for i := floor(sx1) to ceiling(sx2)-1 do //horizontal source pixels
        begin
         dx := 1;
         if sx1 > i then begin
                         dx := dx-(sx1-i);
                        end;
         if sx2 < i+1 then begin
                           dx := dx-(i+1-sx2);
                          end;
         color := sbm.XPixel[srt.left+i,srt.top+j];
         sB := color and $ff;
         sG := (color shr 8) and $ff;
         sR := (color shr 16) and $ff;
         AP := dx*dy*fi2;
         destR := destR + sR*AP;
         destG := destG + sG*AP;
         destB := destB + sB*AP;
        end;//for i
      end;//for j
      color := trunc(destB);
      color := color or (trunc(destG) shl 8);
      color := color or (trunc(destR) shl 16);
      Xpixel[drt.left+x,drt.top+y] := color;
    end;//for x
  end;//for y
end;

//-------- line,pen,brush properties

procedure TXBitmap.setXLineStyle(ps : byte);
begin
 if ps >= 20 then ps := 0;
 FXlinePattern := linepattern[ps];
end;

procedure TXBitmap.setXPencolor(color : DWORD);
begin
 FXpencolor := (swapcolor(color) and $00fffffc) or FXpenlevel;
end;

procedure TXBitmap.setXPenlevel(lev : byte);
begin
 FXpenlevel := lev and $3;
 FXpenColor := (FXpencolor and $00fffffc) or FXpenlevel;
end;

procedure TXBitmap.setXBrushlevel(lev : byte);
begin
 FXBrushlevel := lev and $3;
 FXbrushColor := (FXbrushcolor and $00fffffc) or FXBrushlevel;
end;

procedure TXBitmap.setXBrushColor(color : DWORD);
begin
 FXBrushcolor := (swapColor(color) and $00fffffc) or FXbrushlevel;
end;

function TXBitmap.getXpencolor : DWORD;
begin
 result := swapColor(FXpencolor);
end;

function TXBitmap.getXbrushColor : DWORD;
begin
 result := swapColor(FXBrushColor);
end;

procedure TXBitmap.setPenwidth(w : byte);
//make pen image
var i,j : byte;
    h,v,r,r2 : single;
    mask : DWORD;
begin
 if w = FXpenwidth then exit;
 if w > 32 then w := 32;
 FXpenwidth := w;
 FXpenBias := (w-1) shr 1;

//------- make pen image --------------------

 for i := 0 to 31 do FXpen[i] := 0;//erase pen
 r := w/2;
 r2 := r * r;
 for j := 0 to (w-1) shr 1 do
  begin
   v := r - (0.5 + j);
   for i := 0 to (w-1) shr 1 do
    begin
     h := r - (0.5 + i);
     if h*h + v*v <= r2 then
      begin
       mask := 1 shl i;
       mask := mask or (1 shl (w-i-1));
       FXpen[j] := FXpen[j] or mask;
       FXpen[w-j-1] := FXpen[w-j-1] or mask;
      end;//if
    end;//for i
  end;//for j
end;

//----------- cliprect ------------------

procedure TXBitmap.setXClipRect(r : Trect);
//check rectangle for proper values , adjust
var h : integer;
begin
 if (r.Left < 0) or (r.left > width) then r.left := 0;
 if (r.Right < 0) or (r.right > width) then r.right := width;
 if (r.top < 0) or (r.top > height) then r.top := 0;
 if (r.bottom < 0) or (r.bottom > height) then r.bottom := height;
 if r.top > r.bottom then
  begin
   h := r.top; r.top := r.bottom; r.bottom := h;
  end;
 if r.left > r.right then
  begin
   h := r.Left; r.left := r.right; r.Right := h;
  end;
 if (r.left = r.right) or (r.top = r.bottom) then
  begin
   r.left := 0; r.right := width;
   r.top := 0; r.bottom := height;
  end;
 FXCliprect := r;
end;

procedure TXBitmap.makeModRect(l,t,r,b : integer; expans : byte);
//called by Xdot,Xline,Xellipse,Xarc,Xfillrect
//make XModRect to indicate last changes
//expans is extra width,height if arrows, fat lines..
begin
 with FXModRect do
  begin
   if l < r then begin
                  left := l - expans; right := r + expans;
                 end
    else begin
          left := r - expans; right := l + expans;
         end;
   if t < b then begin
                  top := t - expans; bottom := b + expans;
                 end
    else begin
          top := b - expans; bottom := t + expans;
         end;
  end;

  FXModRect := InterRect(FXModRect,FXClipRect);

end;

//--------- fillstyle ------------------

procedure TXBitmap.setXFillStyle(fst : byte);
//generate both 8*8 and 24*24 fillmaps
//select pattern fst = 0..15
const p8 : word = $111;
      p9 : word = $888;
     
var i : byte;
begin
 if fst = FXFillstyle then exit;
//
 FXFillStyle := fst;

//------- 8 * 8 fill pattern

 for i := 0 to 7 do FXfillMap8[i] := 0;//transparent
 case fst of
   0 : for i := 0 to 7 do
        case i of
                     0,7 : FXfillMap8[i] := $81;
                     1,6 : FXfillMap8[i] := $42;
                     2,5 : FXfillMap8[i] := $3c;
                     3,4 : FXfillMap8[i] := $24;
        end;
  1 : for i := 0 to 7 do FXfillMap8[i] := $01 shl i;
  2 : for i := 0 to 7 do FXfillMap8[i] := $80 shr i;
  3 : begin
       for i := 0 to 7 do FXfillMap8[i] := $1 shl i;
       for i := 0 to 7 do FXfillMap8[i] := FXfillMap8[i] or ($80 shr i);
      end;
  4 : FXfillMap8[0] := $ff;
  5 : for i := 0 to 7 do FXfillMap8[i] := $80;
  6 : begin
       FXfillMap8[0] := $ff;
       for i := 0 to 7 do FXfillMap8[i] := FXfillMap8[i] or $80;
      end;
  7 : for i := 0 to 7 do
       if (i and 1 <> 0) then FXfillMap8[i] := $aa else FXfillMap8[i] := $55;
  8 : for i := 0 to 7 do FXfillMap8[i] := ((p8 shl i) and $ff0) shr 4;
  9 : for i := 0 to 7 do FXfillMap8[i] := (p9 shr i) and $ff;
 10 : for i := 0 to 7 do
       begin
        FXfillMap8[i] := ((p8 shl i) and $ff0) shr 4;
        FXfillMap8[i] := FXfillMap8[i] or ((p9 shr i) and $ff);
       end;
  11 : for i := 0 to 7 do
        if i and $3 = 0 then FXfillMap8[i] := $ff;
  12 : for i := 0 to 7 do FXfillMap8[i] := $88;
  13 : for i := 0 to 7 do
        begin
         if i and $3 = 0 then FXfillMap8[i] := $ff;
         FXfillMap8[i] := FXfillMap8[i] or $88;
        end;
  14 : for i := 0 to 7 do
        case i and $3 of
                     0,2 : FXfillMap8[i] := $44;
                       1 : FXfillMap8[i] := $aa;
        end;
  15 : for i := 0 to 7 do FXFillMap8[i] := $ff;
 end;//case

//------- 24 * 24 fill pattern -------------

 for i := 0 to 23 do FXfillMap24[i] := 0;
 case fst of
  0 : for i := 0 to 23 do
        case i of
                     0,23 : FXfillMap24[i] := $c00003;
                     1,22 : FXfillMap24[i] := $600006;
                     2,21 : FXfillMap24[i] := $30000c;
                     3,20 : FXfillMap24[i] := $180018;
                     4,19 : FXfillMap24[i] := $0c0030;
                     5,18 : FXfillMap24[i] := $060060;
                     6,7,16,17 : FXfillMap24[i] := $07ffe0;
                     8..15: FXfillmap24[i] := $060060;
        end;
  1 : begin
       for i := 0 to 22 do FXfillMap24[i] := $3 shl i;
       FxfillMap24[23] := $800001;
      end;
  2 : begin
       for i := 0 to 22 do FXfillMap24[i] := $c00000 shr i;
       Fxfillmap24[23] := $800001;
      end;
  3 : for i := 0 to 22 do begin
                           FXfillMap24[i] := $3 shl i;
                           FXfillMap24[i] := FXfillMap24[i] or ($c00000 shr i);
                           FXfillmap24[23] := $800001;
                          end;
  4 : begin
       FXfillMap24[0] := $ffffff;
       FXfillMap24[1] := $ffffff;
      end;
  5 : for i := 0 to 23 do FXfillMap24[i] := $c00000;
  6 : begin
       FXfillMap24[0] := $ffffff;
       FXfillMap24[1] := $ffffff;
       for i := 0 to 23 do FXfillMap24[i] := FXfillMap24[i] or $c00000;
      end;
  7 : for i := 0 to 7 do
       if (i and 1 <> 0) then
        begin
         FXfillMap24[3*i] := $145145;
         FXfillMap24[3*i+1] := $082082;
         FXfillMap24[3*i+2] := $145145;
        end
       else begin
             FXfillmap24[3*i] := $a28a28;
             FXfillMap24[3*i+1] := $410410;
             FXfillmap24[3*i+2] := $a28a28;
            end;
  9 : begin
       for i := 0 to 10 do
        begin
         FXfillMap24[i] := $c00c00 shr i;
         FXfillMap24[i+12] := FXfillMap24[i];
        end;
        FXfillMap24[11] := $801801;
        FXfillmap24[23] := $801801;
       end;
  8 : begin
       for i := 0 to 10 do
        begin
         FXfillMap24[i] := $003003 shl i;
         FXfillMap24[i+12] := FXFillmap24[i];
        end;
        FXfillMap24[11] := $801801;
        FXFillMap24[23] := $801801;
       end;
 10 : begin
       for i := 0 to 10 do
        begin
         FXfillMap24[i] := $c00c00 shr i;
         FXfillMap24[i] := FXFillMap24[i] or ($003003 shl i);
         FXfillMap24[i+12] := FXfillMap24[i];
        end;
        FXfillMap24[11] := $801801;
        FXfillmap24[23] := $801801;
       end;
  11 : for i := 0 to 23 do
        if (i mod 12) and $fe = 0 then FXfillMap24[i] := $ffffff;
  12 : for i := 0 to 23 do FXfillMap24[i] := $c00c00;
  13 : for i := 0 to 23 do
        begin
         if (i mod 12) and $fe = 0 then FXfillMap24[i] := $ffffff;
         FXfillMap24[i] := FXfillMap24[i] or $c00c00;
        end;
  14 : for i := 0 to 23 do
        case i mod 12 of
             0,8 : FXfillMap24[i] := $080080;
             1,7 : FXfillMap24[i] := $1c01c0;
             2,6 : FXfillMap24[i] := $360360;
             3,5 : FXfillMap24[i] := $630630;
               4 : FXFillMap24[i] := $c1fc1f;
         9,10,11 : FXFillMap24[i] := $080080;
        end;//case
  15 : for i := 0 to 23 do
        if (i and 1) = 0 then FXfillMap24[i] := $555555
        else FXfillMap24[i] := $aaaaaa;
 end;//case
end;

//------------- single pixel acces -------------------

function TXBitmap.getPixPos(x,y : integer):DWORD;
//return address of [x,y] pixel in DWORD format
//for use:  pointer := PDW(dword);
begin
 result := FXpbase - FXlineStep*y + (x shl 2);
end;

function TXBitmap.getXpixel(x,y : integer): DWORD;
//read single pixel from (x,y)
//format is 32 bit, no conversion
var pLine : DWORD;
    p : PDW;
begin
 result := $ffffff;
 if (y < 0) or (y >= height) then exit;
 pline := FXpbase-FXlineStep*y;
 if (x >= 0) and (x < width) then
  begin
   p := PDW(pline + x shl 2);
   result := p^;
  end;
end;

procedure TXBitmap.setXpixel(x,y : integer; dw : DWORD);
//write  pixel from (x,y) , fast access , ignore xcliprect
//dw is 32 bit color, no conversion
var pline : DWORD;
    p : PDW;
begin
 if (y < 0) or (y >= height) then exit;
 pline := FXpbase-FXlineStep*y;
 if (x >= 0) and (x < width) then
  begin
   p := PDW(pline + x shl 2);
   p^ := dw;
  end;
end;

//------------- support procedures

function UniRect(var r1,r2 : TRect) : Trect;
//r1,r2 must be proper rects : left < right, top < bottom
//union of rectangles destRect,sourceRect
begin
 with result do
  begin
   if r1.left < r2.Left then left := r1.left else left := r2.left;
   if r1.right > r2.right then right := r1.right else right := r2.right;
   if r1.top < r2.top then top := r1.top else top := r2.top;
   if r1.bottom > r2.bottom then bottom := r1.bottom else bottom := r2.bottom;
  end;
end;

function InterRect(var r1,r2 : Trect) : Trect;
begin
 with result do
  begin
   if r1.left > r2.Left then left := r1.left else left := r2.left;
   if r1.right < r2.right then right := r1.right else right := r2.right;
   if r1.top > r2.top then top := r1.top else top := r2.top;
   if r1.bottom < r2.bottom then bottom := r1.bottom else bottom := r2.bottom;
  end;
end;

function PackRect(x1,y1,x2,y2 : integer) : Trect;
//pack x1,y1,x2,y2 in rectangle , sort
begin
 with result do
  begin
   if x1 < x2 then begin
                      left := x1; right := x2;
                     end
    else begin
          left := x2; right := x1;
         end;
   if y1 < y2 then begin
                      top := y1; bottom := y2;
                     end
    else begin
          top := y2; bottom := y1;
         end;
 end;//with
end;

function swapColor(c : DWORD) : DWORD;
//convert 24 bit color to 32 or 32 to 24
//trade blue and red fields
//reset bits 24..31
begin
 result := (c shr 16) and $ff;
 result := result or (c and $0000ff00);
 result := result or ((c and $ff) shl 16);
end;

end.